(* Persistent queue *)

type 'a t = 'a list * int * 'a list lazy_t * int * 'a list

exception Empty;;

let empty = ([], 0, lazy [], 0, []);;
let is_empty (_, lf, _, _, _) = lf = 0;;
let checkw = function
  | ([], lf, f, lr, r) -> (Lazy.force f, lf, f, lr, r)
  | q -> q
;;
let check ((w, lf, f, lr, r) as q) =
  if lr < lf then
    checkw q
  else
    let f' = Lazy.force f in
    checkw (f', lf + lr, lazy (f' @ List.rev r), 0, [])
;;
let add x (w, lf, f, lr, r) = check (w, lf, f, lr + 1, x :: r);;
let head = function
  | ([],     _, _, _, _) -> raise Empty
  | (x :: _, _, _, _, _) -> x
;;
let tail = function
  | ([],      _, _,  _, _) -> raise Empty
  | (_ :: w, lf, f, lr, r) -> check (w, lf - 1, lazy (List.tl (Lazy.force f)), lr, r)
;;
let take q = (head q, tail q);;
